"""
Defines how to normalize a focusing reflectometry dataset by a reference measurement.
"""
import logging
import os
import numpy as np
from typing import List, Optional

from orsopy import fileio

from .event_data_types import EventDatasetProtocol
from .header import Header
from .options import NormalisationMethod
from .instrument import Detector, LZGrid


class LZNormalisation:
    file_list = List[str]
    angle: float
    monitor: float
    norm: np.ndarray

    def __init__(self, reference:EventDatasetProtocol, normalisationMethod: NormalisationMethod, grid: LZGrid):
        self.angle = reference.geometry.nu-reference.geometry.mu
        lamda_e = reference.data.events.lamda
        detZ_e = reference.data.events.detZ
        self.monitor = np.sum(reference.data.pulses.monitor)
        norm_lz, _, _ = np.histogram2d(lamda_e, detZ_e, bins=(grid.lamda(), grid.z()))
        if normalisationMethod==NormalisationMethod.direct_beam:
            self.norm = np.flip(norm_lz, 1)
        else:
            # correct for reference sm reflectivity
            lamda_l = grid.lamda()
            theta_z = self.angle+Detector.delta_z
            lamda_lz = (grid.lz().T*lamda_l[:-1]).T
            theta_lz = grid.lz()*theta_z
            qz_lz = 4.0*np.pi*np.sin(np.deg2rad(theta_lz))/lamda_lz
            # TODO: introduce variable for `m` and propably for the slope
            # Correct reflectivity of m=5 supermirror
            Rsm_lz = np.ones(np.shape(qz_lz))
            Rsm_lz = np.where(qz_lz>0.0217, 1-(qz_lz-0.0217)*(0.0625/0.0217), Rsm_lz)
            Rsm_lz = np.where(qz_lz>0.0217*5, np.nan, Rsm_lz)
            self.norm = norm_lz/Rsm_lz
        self.file_list = [os.path.basename(entry) for entry in reference.file_list]

    @classmethod
    def from_file(cls, filename, check_hash=None) -> Optional['LZNormalisation']:
        self = super().__new__(cls)
        with open(filename, 'rb') as fh:
            hash = str(np.load(fh, allow_pickle=True))
            self.file_list = np.load(fh, allow_pickle=True).tolist()
            self.angle = np.load(fh, allow_pickle=True)
            self.norm = np.load(fh, allow_pickle=True)
            self.monitor = np.load(fh, allow_pickle=True)
        if check_hash is not None and hash != check_hash:
            logging.info('    file hash does not match this reduction configuration')
            raise ValueError('file hash does not match this reduction configuration')
        return self

    @classmethod
    def unity(cls, grid:LZGrid) -> 'LZNormalisation':
        logging.warning(f'normalisation is unity')
        self = super().__new__(cls)
        self.norm = grid.lz()
        self.file_list = []
        self.angle = 1.
        self.monitor = 1.
        return self

    @classmethod
    def model(cls, grid:LZGrid) -> 'LZNormalisation':
        # generate a normalization based on angular and wavelength distribution model
        # TODO: add options for sample size for better absolute normalization
        logging.warning(f'normalisation is model')
        self = super().__new__(cls)
        self.angle = 1.0
        self.monitor = 4e6

        lamda_l  = grid.lamda()
        lamda_c = (lamda_l[:-1]+lamda_l[1:])/2

        delta = np.rad2deg(np.arctan2(grid.z(), Detector.distance))/2.0
        delta_c = (delta[:-1]+delta[1:])/2-delta.mean()

        # approximate spectrum by Maxwell-Boltzmann and intensity by linear footprint
        a = 3.8
        Ilambda = np.sqrt(2./np.pi)*lamda_c**2/a**3*np.exp(-lamda_c**2/(2.*a**2))
        Idelta = np.where(abs(delta_c)<0.75, (self.angle-delta_c), np.nan)

        self.norm = 1e6*Ilambda[:, np.newaxis]*Idelta[np.newaxis, :]

        return self

    def safe(self, filename, hash):
        with open(filename, 'wb') as fh:
            np.save(fh, hash, allow_pickle=False)
            np.save(fh, np.array(self.file_list), allow_pickle=False)
            np.save(fh, np.array(self.angle), allow_pickle=False)
            np.save(fh, self.norm, allow_pickle=False)
            np.save(fh, self.monitor, allow_pickle=False)

    def update_header(self, header:Header):
        header.measurement_additional_files = [fileio.File(file=os.path.basename(entry)) for entry in self.file_list]

    def smooth(self, width):
        logging.debug(f'apply convolution with gaussian along lambda axis to smooth norm, sigma={width}')
        try:
            from scipy.signal import fftconvolve
        except ImportError:
            self._smooth_slow(width)
        kx = np.arange(self.norm.shape[0])-self.norm.shape[0]/2.
        kernel = np.exp(-0.5*kx**2/width**2)
        kernel/=kernel.sum()
        kernel = kernel[:, np.newaxis]*np.ones(self.norm.shape[1])[np.newaxis, :]
        unorm = np.where(np.isnan(self.norm), 0., self.norm)
        nnorm = fftconvolve(unorm, kernel, mode='same', axes=0)
        nnorm[np.isnan(self.norm)] = np.nan
        self.norm = nnorm

    def _smooth_slow(self, width):
        # like smooth but using numpy buildin slow convolve
        nnorm = np.zeros_like(self.norm)

        kx = np.arange(self.norm.shape[0])-self.norm.shape[0]/2.
        kernel = np.exp(-0.5*kx**2/width**2)
        kernel/=kernel.sum()
        unorm = np.where(np.isnan(self.norm), 0., self.norm)

        for row in range(self.norm.shape[1]):
            nnorm[:, row] = np.convolve(unorm[:, row], kernel, mode='same')
        nnorm[np.isnan(self.norm)] = np.nan
        self.norm = nnorm



"""
amor-nicos vizualising data from Amor@SINQ, PSI

Author: Jochen Stahn (algorithms, python draft),
        Artur Glavic (structuring and optimisation of code)
"""
import logging

# need to do absolute import here as pyinstaller requires it
from eos.options import E2HConfig, ReaderConfig, ExperimentConfig, E2HReductionConfig
from eos.command_line import commandLineArgs
from eos.logconfig import setup_logging, update_loglevel


def main():
    setup_logging()
    logging.getLogger('matplotlib').setLevel(logging.WARNING)

    # read command line arguments and generate classes holding configuration parameters
    clas = commandLineArgs([ReaderConfig, ExperimentConfig, E2HReductionConfig],
                           'amor-nicos')
    update_loglevel(clas.verbose)
    if clas.verbose<2:
        # only log info level in logfile
        logger = logging.getLogger()  # logging.getLogger('quicknxs')
        logger.setLevel(logging.INFO)

    reader_config = ReaderConfig.from_args(clas)
    experiment_config = ExperimentConfig.from_args(clas)
    reduction_config = E2HReductionConfig.from_args(clas)
    config = E2HConfig(reader_config, experiment_config, reduction_config)

    logging.warning('######## amor-nicos - Nicos histogram for Amor ########')
    from eos.reduction_kafka import KafkaReduction

    # only import heavy module if sufficient command line parameters were provided
    from eos.reduction_reflectivity import ReflectivityReduction
    # Create reducer with these arguments
    reducer = KafkaReduction(config)
    # Perform actual reduction
    reducer.reduce()

    logging.info('######## amor-nicos - finished ########')

if __name__ == '__main__':
    main()



"""
Functions used to directly access information from nicos.
"""

import socket
import platform
import logging
import subprocess

ON_AMOR = platform.node().startswith('amor')
NICOS_CACHE_DIR = '/home/data/nicosdata/cache/'
GREP = '/usr/bin/grep "value"'


def lookup_nicos_value(key, nicos_key, dtype=float, suffix='', year="2026"):
    # TODO: Implement direct communication to nicos to read the cache
    if nicos_key=='ignore':
        return dtype(0)
    try:
        logging.debug(f'     trying socket request for device {nicos_key}')
        response = nicos_single_request(nicos_key)
        logging.info(f"     using parameter {nicos_key} from nicos cache via socket")
        return dtype(response)
    except Exception as e:
        logging.debug(f'       socket request failed with {e!r}')
    if ON_AMOR:
        logging.debug(f"     trying to extract {nicos_key} from nicos cache files")
        call = f'{GREP} {NICOS_CACHE_DIR}nicos-{nicos_key}/{year}{suffix}'
        try:
            value = str(subprocess.getoutput(call)).split('\t')[-1]
            logging.info(f"     using parameter {nicos_key} from nicos cache file")
            return dtype(value)
        except Exception:
            logging.error(f"       couldn't get value from nicos cache {nicos_key}, {call}")
            return dtype(0)
    else:
        logging.warning(f"     parameter {key} not found, relpace by zero")
        return dtype(0)

def nicos_single_request(device):
    sentinel = b'\n'
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.settimeout(1.0)
        s.connect(('amor', 14869))

        tosend = f'@nicos/{device}/value?\n'

        # write request
        # self.log.debug("get_explicit: sending %r", tosend)
        s.sendall(tosend.encode())

        # read response
        data = b''
        while not data.endswith(sentinel):
            newdata = s.recv(8192)  # blocking read
            if not newdata:
                raise IOError('cache closed connection')
            data += newdata
        s.shutdown(socket.SHUT_RDWR)
    return data.decode('utf-8').split('=')[-1]



"""
Classes describing the AMOR instrument configuration used during reduction.
"""

import logging
import numpy as np

from . import const

try:
    from functools import cache
except ImportError:
    # python <3.9
    def cache(func): return func


class Detector:
    nBlades  = 14  # number of active blades in the detector
    nWires   = 32  # number of wires per blade
    nStripes = 64  # number of stipes per blade
    angle    = np.deg2rad(5.1)  # deg  angle of incidence of the beam on the blades (def: 5.1)
    dZ       = 4.0*np.sin(angle)  # mm  height-distance of neighboring pixels on one blade
    dX       = 4.0*np.cos(angle)  # mm  depth-distance of neighboring pixels on one blace
    bladeZ   = 10.455  # mm  distance between detector blades
    zero     = 0.5*nBlades*bladeZ  # mm  vertical center of the detector
    distance = 4000.  # mm  distance from focal point to leading blade edge

    delta_z: np.ndarray
    pixelLookUp: np.ndarray

    @staticmethod
    def resolve_pixels():
        """
        Determine spatial coordinats and angles from pixel number,
        does only have to be computed once for the detector
        """
        if hasattr(Detector, 'pixelLookUp'):
            return
        nPixel = Detector.nWires * Detector.nStripes * Detector.nBlades
        pixelID = np.arange(nPixel)
        (bladeNr, bPixel) = np.divmod(pixelID, Detector.nWires * Detector.nStripes)
        (bZi, detYi)      = np.divmod(bPixel, Detector.nStripes)                     # z index on blade, y index on detector
        detZi             = bladeNr * Detector.nWires + bZi                          # z index on detector
        detX              = bZi * Detector.dX                                        # x position in detector
        # detZ              = Detector.zero - bladeNr * Detector.bladeZ - bZi * Detector.dZ      # z position on detector
        bladeAngle        = np.rad2deg( 2. * np.arcsin(0.5*Detector.bladeZ / Detector.distance) )
        delta             = (Detector.nBlades/2. - bladeNr) * bladeAngle \
                            - np.rad2deg( np.arctan(bZi*Detector.dZ / ( Detector.distance + bZi * Detector.dX) ) )
        delta_z      = delta[detYi==1]
        pixel_lookup=np.vstack((detYi.T, detZi.T, detX.T, delta.T)).T
        Detector.delta_z = delta_z
        Detector.pixelLookUp = pixel_lookup

# guarantee that pixelLookUp has been computed
Detector.resolve_pixels()

class LZGrid:
    dldl = 0.005  # Delta lambda / lambda

    # as using cahced results, make sure the object is not modified
    @property
    def qResolution(self):
        return self._qResolution
    @property
    def qzRange(self):
        return self._qzRange

    def __init__(self, qResolution, qzRange, lambda_overwrite=None):
        self._qResolution = qResolution
        self._qzRange = qzRange
        if lambda_overwrite is None:
            self.lamdaMax = const.lamdaMax
            self.lamdaCut = const.lamdaCut
        else:
            self.lamdaCut, self.lamdaMax = lambda_overwrite

    @property
    @cache
    def shape(self):
        # gives the shape of the grid, not of the bin-edges
        return (self.lamda().shape[0]-1, self.z().shape[0]-1)

    @cache
    def q(self):
        resolutions = [0.005, 0.01, 0.02, 0.025, 0.04, 0.05, 0.1, 1]
        a, b = np.histogram([self.qResolution], bins = resolutions)
        dqdq = np.matmul(b[:-1],a)
        if dqdq != self.qResolution:
            logging.info(f'#   changed resolution to {dqdq}')
        qq = 0.01
        # linear up to qq
        q_grid = np.arange(0, qq, qq*dqdq)
        # exponential from qq on
        q_grid = np.append(q_grid, qq*(1.+dqdq)**np.arange(int(np.log(self.qzRange[1]/qq)/np.log(1+dqdq))))
        q_grid = q_grid[q_grid>=self.qzRange[0]]
        return q_grid

    @cache
    def lamda(self):
        lamdaMax = self.lamdaMax
        lamdaMin = self.lamdaCut
        lamda_grid = lamdaMin*(1+self.dldl)**np.arange(int(np.log(lamdaMax/lamdaMin)/np.log(1+self.dldl)+1))
        return lamda_grid

    @cache
    def z(self):
        # TODO: shouldn't this be -0.5 to be the edges of each pixel?
        return np.arange(Detector.nBlades*Detector.nWires+1)

    @cache
    def lz(self):
        return np.ones(( self.lamda().shape[0]-1, self.z().shape[0]-1))

    @cache
    def delta(self, detectorDistance):
        # unused for now
        bladeAngle = np.rad2deg( 2. * np.arcsin(0.5*Detector.bladeZ / detectorDistance) )
        blade_grid = np.arctan( np.arange(33) * Detector.dZ / ( detectorDistance + np.arange(33) * Detector.dX) )
        blade_grid = np.rad2deg(blade_grid)
        stepWidth  = blade_grid[1] - blade_grid[0]
        blade_grid = blade_grid - 0.2 * stepWidth

        delta_grid = []
        for b in np.arange(Detector.nBlades-1):
            delta_grid = np.concatenate((delta_grid, blade_grid), axis=None)
            blade_grid = blade_grid + bladeAngle
            delta_grid = delta_grid[delta_grid<blade_grid[0]-0.5*stepWidth]
        delta_grid = np.concatenate((delta_grid, blade_grid), axis=None)

        return -np.flip(delta_grid) + 0.5*Detector.nBlades * bladeAngle



import os
import numpy as np
import logging

from unittest import TestCase
from datetime import datetime
from copy import deepcopy

from orsopy.fileio import Person, Experiment, Sample, InstrumentSettings, Value, ValueRange, Polarization

from eos import const
from eos.header import Header
from eos.event_data_types import EVENT_BITMASKS, AmorGeometry, AmorTiming, AmorEventStream, \
    EventDataAction, EventDatasetProtocol, PACKET_TYPE, PC_TYPE, PULSE_TYPE, EVENT_TYPE, append_fields
from eos.event_handling import ApplyPhaseOffset, ApplyParameterOverwrites, CorrectChopperPhase, CorrectSeriesTime, \
    AssociatePulseWithMonitor, FilterMonitorThreshold, FilterStrangeTimes, TofTimeCorrection, ApplyMask
from eos.event_analysis import ExtractWalltime, MergeFrames, AnalyzePixelIDs, CalculateWavelength, CalculateQ, \
    FilterQzRange, FilterByLog
from eos.options import MonitorType, IncidentAngle, ExperimentConfig


class MockEventData:
    """
    Simulated dataset to be used with event handling unit tests
    """
    geometry: AmorGeometry
    timing: AmorTiming
    data: AmorEventStream

    def __init__(self):
        self.geometry = AmorGeometry(mu=2.0, nu=1.0, kap=0.5, kad=0.0, div=1.5,
                                     chopperSeparation=1000.0, detectorDistance=4000., chopperDetectorDistance=18842.)
        self.timing = AmorTiming(
                ch1TriggerPhase=-9.1, ch2TriggerPhase=6.75,
                chopperPhase=0.17, chopperSpeed=500., tau=0.06
                )
        self.create_data()

    def create_data(self):
        # list of events, here with random time of fligh and pixel location
        events = np.recarray((10000, ), dtype=EVENT_TYPE)
        events.tof = np.random.uniform(low=0., high=0.12, size=events.shape)
        events.pixelID = np.random.randint(0, 28671, size=events.shape)
        events.mask = 0

        # list of data packates containing previous events
        packets = np.recarray((1000,), dtype=PACKET_TYPE)
        packets.start_index = np.linspace(0, events.shape[0]-1, packets.shape[0], dtype=np.uint32)
        packets.time = np.linspace(1700000000000000000, 1700000000000000000+3_600_000_000,
                                   packets.shape[0], dtype=np.int64)

        # chopper pulses within the measurement time
        pulses = np.recarray((packets.shape[0],), dtype=PULSE_TYPE)
        pulses.monitor = 1.0
        pulses.time = packets.time

        # proton current information with independent timing
        proton_current = np.recarray((50,), dtype=PC_TYPE)
        proton_current.current = 1500.0
        proton_current[np.random.randint(0, proton_current.shape[0]-1, 10)] = 0. # random time with no current
        proton_current.time = np.linspace(1700000000000000300, 1700000000000000000+3_600_000_000,
                                   proton_current.shape[0], dtype=np.int64)

        self.data = AmorEventStream(events, packets, pulses, proton_current)
        self.orig_data = deepcopy(self.data)


    def append(self, other):
        raise NotImplementedError("Just for testing, no append")

    def update_header(self, header:Header):
        # update a header with the information read from file
        header.owner = Person(name="test user", affiliation='PSI')
        header.experiment = Experiment(title='test experiment', instrument='amor',
                                       start_date=datetime.now(), probe="neutron")
        header.sample = Sample(name='test sample')
        header.measurement_instrument_settings = InstrumentSettings(incident_angle=Value(1.5, 'deg'),
            wavelength = ValueRange(3.0, 12.5, 'angstrom'),
            polarization = Polarization.unpolarized)

    def update_info_from_logs(self):
        RELEVANT_ITEMS = ['sample_temperature', 'sample_magnetic_field', 'polarization_config_label']
        for key, log in self.data.device_logs.items():
            if key not in RELEVANT_ITEMS:
                continue
            if log.value.dtype in [np.int8, np.int16, np.int32, np.int64]:
                # for integer items (flags) report the most common one
                value = np.bincount(log.value).argmax()
                if logging.getLogger().getEffectiveLevel() <= logging.DEBUG \
                        and np.unique(log.value).shape[0]>1:
                    logging.debug(f'        filtered values for {key} not unique, '
                                  f'has {np.unique(log.value).shape[0]} values')
            else:
                value = log.value.mean()
            if key == 'polarization_config_label':
                self.instrument_settings.polarization = Polarization(const.polarizationConfigs[value])
            elif key == 'sample_temperature':
                self.sample.sample_parameters['temperature'].magnitue = value
            elif key == 'sample_magnetic_field':
                self.sample.sample_parameters['magnetic_field'].magnitue = value


class TestActionClass(TestCase):
    @classmethod
    def setUpClass(cls):
        """
        Create test classes to be used
        """
        class T1(EventDataAction):
            def perform_action(self, event: EventDatasetProtocol):
                event.data.events.mask += 1
        class T2(EventDataAction):
            def perform_action(self, event: EventDatasetProtocol):
                event.data.events.mask += 2
        class T4(EventDataAction):
            def perform_action(self, event: EventDatasetProtocol):
                event.data.events.mask += 4
        cls.T1=T1; cls.T2=T2; cls.T4=T4

        class H1(EventDataAction):
            def perform_action(self, event: EventDatasetProtocol):
                ...
            def update_header(self, header:Header) ->None:
                header.sample.name = 'h1'
        class H2(EventDataAction):
            def perform_action(self, event: EventDatasetProtocol):
                ...
            def update_header(self, header: Header) -> None:
                header.sample.name = 'h2'
        class HN(EventDataAction):
            def __init__(self, n):
                self._n = n
            def perform_action(self, event: EventDatasetProtocol):
                ...
            def update_header(self, header: Header) -> None:
                header.sample.name = self._n
        cls.H1=H1; cls.H2=H2; cls.HN = HN

    def setUp(self):
        self.d = MockEventData()
        self.header = Header()
        self.d.update_header(self.header)

    def test_individual(self):
        t1 = self.T1()
        t2 = self.T2()
        t4 = self.T4()

        np.testing.assert_array_equal(self.d.data.events.mask, 0)
        t1.perform_action(self.d)
        np.testing.assert_array_equal(self.d.data.events.mask, 1)
        t2.perform_action(self.d)
        np.testing.assert_array_equal(self.d.data.events.mask, 3)
        t4.perform_action(self.d)
        np.testing.assert_array_equal(self.d.data.events.mask, 7)

    def test_header(self):
        h1 = self.H1()
        h2 = self.H2()
        h3 = self.HN('h3')
        h4 = self.HN('h4')

        h1.update_header(self.header)
        self.assertEqual(self.header.sample.name, 'h1')
        h2.update_header(self.header)
        self.assertEqual(self.header.sample.name, 'h2')
        h3.update_header(self.header)
        self.assertEqual(self.header.sample.name, 'h3')
        h4.update_header(self.header)
        self.assertEqual(self.header.sample.name, 'h4')

    def test_combination(self):
        t1 = self.T1()
        t2 = self.T2()
        t4 = self.T4()
        t12 = t1 | t2
        t24 = t2 | t4
        t1224 = t1 | t2 | t2 | t4
        t1224b = t12 | t24

        np.testing.assert_array_equal(self.d.data.events.mask, 0)
        t12.perform_action(self.d)
        np.testing.assert_array_equal(self.d.data.events.mask, 3)
        t24.perform_action(self.d)
        np.testing.assert_array_equal(self.d.data.events.mask, 9)

        t1224.perform_action(self.d)
        np.testing.assert_array_equal(self.d.data.events.mask, 18)
        t1224b.perform_action(self.d)
        np.testing.assert_array_equal(self.d.data.events.mask, 27)


    def test_combine_header(self):
        h1 = self.H1()
        h2 = self.H2()
        h3 = self.HN('h3')
        h4 = self.HN('h4')

        (h1|h2).update_header(self.header)
        self.assertEqual(self.header.sample.name, 'h2')
        (h2|h1).update_header(self.header)
        self.assertEqual(self.header.sample.name, 'h1')
        (h3|h4).update_header(self.header)
        self.assertEqual(self.header.sample.name, 'h4')
        (h4|h3).update_header(self.header)
        self.assertEqual(self.header.sample.name, 'h3')

    def test_abstract_misssing(self):
        with self.assertRaises(TypeError):
            class E(EventDataAction):
                ...
            _ = E()

    def test_hash(self):
        """
        Check that hashes of different actions are different but
        instances of same action have same hash
        """
        t1 = self.T1()
        t1b = self.T1()
        t2 = self.T2()
        t4 = self.T4()
        h3 = self.HN('h3')
        h3b = self.HN('h3')
        h4 = self.HN('h4')

        self.assertNotEqual(t1.action_hash(), t2.action_hash())
        self.assertNotEqual(t2.action_hash(), t4.action_hash())
        self.assertNotEqual(t1.action_hash(), t4.action_hash())
        self.assertNotEqual(h3.action_hash(), h4.action_hash())
        self.assertEqual(t1.action_hash(), t1b.action_hash())
        self.assertEqual(h3.action_hash(), h3b.action_hash())


class TestSimpleActions(TestCase):
    def setUp(self):
        self.d = MockEventData()
        self.header = Header()
        self.d.update_header(self.header)

    def test_chopper_phase(self):
        cp = CorrectChopperPhase()
        cp.perform_action(self.d)
        np.testing.assert_array_equal(
                self.d.data.events.tof,
                self.d.orig_data.events.tof+
                self.d.timing.tau*(self.d.timing.ch1TriggerPhase-self.d.timing.chopperPhase/2)/180
                )

    def _extract_walltime(self):
        # Extract wall time for events and orig copy
        wt = ExtractWalltime()
        d = self.d.data
        self.d.data = self.d.orig_data
        wt.perform_action(self.d)
        self.d.data = d
        wt.perform_action(self.d)

    def test_extract_walltime(self):
        self._extract_walltime()
        # wallTime should be always a time present in the packet times
        np.testing.assert_array_equal(np.isin(self.d.data.events.wallTime, self.d.data.packets.time), True)
        # make sure extraction works on both original and copy
        np.testing.assert_array_equal(self.d.data.events.wallTime, self.d.orig_data.events.wallTime)

    def test_series_time(self):
        corr = 100
        ct = CorrectSeriesTime(corr)

        with self.assertRaises(ValueError):
            ct.perform_action(self.d)

        self._extract_walltime()


        ct.perform_action(self.d)
        np.testing.assert_array_equal(
                self.d.data.pulses.time,
                self.d.orig_data.pulses.time-corr
                )
        np.testing.assert_array_equal(
                self.d.data.events.wallTime,
                self.d.orig_data.events.wallTime-corr
                )
        np.testing.assert_array_equal(
                self.d.data.proton_current.time,
                self.d.orig_data.proton_current.time-corr
                )

    def test_associate_monitor(self):
        amPC = AssociatePulseWithMonitor(MonitorType.proton_charge)
        amT = AssociatePulseWithMonitor(MonitorType.time)
        amN = AssociatePulseWithMonitor(MonitorType.neutron_monitor)

        self.d.data.pulses.monitor = 13
        amN.perform_action(self.d)
        np.testing.assert_array_equal(self.d.data.pulses.monitor, 1)

        self.d.data.pulses.monitor = 13
        amT.perform_action(self.d)
        np.testing.assert_array_equal(self.d.data.pulses.monitor, np.float32(2*self.d.timing.tau))

        self.d.data.pulses.monitor = 13
        amPC.perform_action(self.d)
        pcm = self.d.data.proton_current.current *2*self.d.timing.tau*1e-3
        np.testing.assert_array_equal(np.isin(self.d.data.pulses.monitor, pcm), True)

    def test_filter_monitor_threashold(self):
        amPC = AssociatePulseWithMonitor(MonitorType.proton_charge)
        fmt = amPC | FilterMonitorThreshold(1000.)
        fma = amPC | FilterMonitorThreshold(2000.)
        fm0 = amPC | FilterMonitorThreshold(-1.0)

        with self.assertRaises(ValueError):
            fmt.perform_action(self.d)

        self._extract_walltime()
        fm0.perform_action(self.d)
        self.assertEqual(self.d.data.events.mask.sum(), 0)
        fmt.perform_action(self.d)
        # calculate, which events should have 0 monitor
        zero_times = self.d.data.pulses.time[self.d.data.pulses.monitor==0]
        zero_sum = np.isin(self.d.data.events.wallTime, zero_times).sum()
        self.assertEqual(self.d.data.events.mask.sum(), zero_sum*EVENT_BITMASKS['MonitorThreshold'])
        # filter all events
        self.d.data.events.mask = 0
        fma.perform_action(self.d)
        self.assertEqual(self.d.data.events.mask.sum(), self.d.data.events.shape[0]*EVENT_BITMASKS['MonitorThreshold'])

    def test_filter_strage_times(self):
        st = FilterStrangeTimes()

        st.perform_action(self.d)
        self.assertEqual(self.d.data.events.mask.sum(), 0)

        # half events should be strange times (outside of ToF frame)
        self.d.data.events.tof += self.d.timing.tau
        st.perform_action(self.d)
        self.assertEqual(self.d.data.events.mask.sum(),
                         (self.d.data.events.tof>2*self.d.timing.tau).sum()*EVENT_BITMASKS['StrangeTimes'])

    def test_apply_phase_offset(self):
        action = ApplyPhaseOffset(12.5)
        action.perform_action(self.d)
        self.assertEqual(self.d.timing.ch1TriggerPhase, 12.5)

    def test_apply_parameter_overwrites(self):
        action = ApplyParameterOverwrites(ExperimentConfig(muOffset=0.25, mu=3.5, nu=4.5))
        action.perform_action(self.d)
        self.assertEqual(self.d.geometry.mu, 3.5)
        self.assertEqual(self.d.geometry.nu, 4.5)

        action = ApplyParameterOverwrites(ExperimentConfig(muOffset=0.25))
        action.perform_action(self.d)
        self.assertEqual(self.d.geometry.mu, 3.75)

        action = ApplyParameterOverwrites(ExperimentConfig(sampleModel='air | Si | Fe'))
        action.update_header(self.header)
        self.assertIsNotNone(self.header.sample.model)

    def test_apply_sample_model_file(self):
        if os.path.isfile('test.yaml'):
            os.remove('test.yaml')
        action = ApplyParameterOverwrites(ExperimentConfig(sampleModel='test.yaml'))
        action.update_header(self.header)
        self.assertIsNone(self.header.sample.model)

        with open('test.yaml', 'w') as fh:
            fh.write("""stack: air | Si | Fe""")

        try:
            action = ApplyParameterOverwrites(ExperimentConfig(sampleModel='test.yaml'))
            action.update_header(self.header)
            self.assertEqual(self.header.sample.model.stack, 'air | Si | Fe')
        finally:
            os.remove('test.yaml')

    def test_tof_time_correction(self):
        action = TofTimeCorrection()
        with self.assertRaises(ValueError):
            action.perform_action(self.d)

        new_events = append_fields(self.d.data.events, [('delta', np.float64)])
        new_events.delta = 10.0
        self.d.data.events = new_events
        tof_before = self.d.data.events.tof.copy()
        action.perform_action(self.d)
        np.testing.assert_allclose(
            self.d.data.events.tof,
            tof_before - (10.0 / 180.0) * self.d.timing.tau
        )

        self.d.create_data()
        new_events = append_fields(self.d.data.events, [('delta', np.float64)])
        new_events.delta = 10.0
        self.d.data.events = new_events
        tof_before = self.d.data.events.tof.copy()
        action = TofTimeCorrection(correct_chopper_opening=False)
        action.perform_action(self.d)
        np.testing.assert_allclose(
            self.d.data.events.tof,
            tof_before - (self.d.geometry.kad / 180.0) * self.d.timing.tau
        )

    def test_apply_mask(self):
        self.d.data.events = self.d.data.events[:6].copy()
        self.d.data.events.mask[:] = [0, 1, 2, 3, 4, 5]

        action = ApplyMask()
        action.perform_action(self.d)
        self.assertEqual(self.d.data.events.shape[0], 1)
        self.assertEqual(self.d.data.events.mask[0], 0)

        self.d.create_data()
        self.d.data.events = self.d.data.events[:6].copy()
        self.d.data.events.mask[:] = [0, 1, 2, 3, 4, 5]
        action = ApplyMask(bitmask_filter=EVENT_BITMASKS['MonitorThreshold'])
        action.perform_action(self.d)
        np.testing.assert_array_equal(self.d.data.events.mask, np.array([0, EVENT_BITMASKS['MonitorThreshold']],
                                                                        dtype=np.int32))

    def test_merge_frames(self):
        action = MergeFrames(lamdaCut=0.0)
        action.perform_action(self.d)
        self.assertEqual(self.d.data.events.tof.shape, self.d.orig_data.events.tof.shape)
        np.testing.assert_array_compare(lambda x,y: x<=y, self.d.data.events.tof, self.d.orig_data.events.tof)
        self.assertTrue((-self.d.timing.tau<=self.d.data.events.tof).all())
        np.testing.assert_array_less(self.d.data.events.tof, self.d.timing.tau)

        action = MergeFrames(lamdaCut=2.0)
        self.d.data.events.tof = self.d.orig_data.events.tof[:]
        action.perform_action(self.d)
        tofCut = 2.0*self.d.geometry.chopperDetectorDistance/const.hdm*1e-13
        self.assertTrue((tofCut-self.d.timing.tau<=self.d.data.events.tof).all())
        self.assertTrue((self.d.data.events.tof<=tofCut+self.d.timing.tau).all())

    def test_analyze_pixel_ids(self):
        action = AnalyzePixelIDs((1000, 1001))
        action.perform_action(self.d)
        self.assertIn('detZ', self.d.data.events.dtype.names)
        self.assertIn('detXdist', self.d.data.events.dtype.names)
        self.assertIn('delta', self.d.data.events.dtype.names)
        self.assertEqual(
            np.bitwise_and(self.d.data.events.mask, EVENT_BITMASKS['yRange']).astype(bool).sum(),
            self.d.data.events.shape[0]
        )
        # TODO: maybe add a test actually checking correct detector-id resolution

    def test_calculate_wavelength(self):
        action = CalculateWavelength((3.0, 5.0))
        with self.assertRaises(ValueError):
            action.perform_action(self.d)

        new_events = append_fields(self.d.data.events, [('detXdist', np.float64)])
        new_events.detXdist = 0.0
        self.d.data.events = new_events
        action.perform_action(self.d)
        self.assertIn('lamda', self.d.data.events.dtype.names)
        flt = self.d.data.events.mask!=EVENT_BITMASKS['LamdaRange']
        # check all wavelength in range not filtered
        np.testing.assert_array_less(self.d.data.events.lamda[flt], 5.0)
        np.testing.assert_array_less(3.0, self.d.data.events.lamda[flt])
        # check all wavelength out of range filtered
        flt = self.d.data.events.mask==EVENT_BITMASKS['LamdaRange']
        self.assertTrue(((self.d.data.events.lamda[flt]<3.0)|(self.d.data.events.lamda[flt]>5.0)).all())

    def test_calculate_q(self):
        action = CalculateQ(IncidentAngle.alphaF)
        with self.assertRaises(ValueError):
            action.perform_action(self.d)

        # TODO: add checks for actual resulting values

        new_events = append_fields(self.d.data.events, [('lamda', np.float64), ('delta', np.float64)])
        new_events.lamda = 5.0
        new_events.delta = 0.0
        self.d.data.events = new_events
        action.perform_action(self.d)
        self.assertIn('qz', self.d.data.events.dtype.names)
        self.assertNotIn('qx', self.d.data.events.dtype.names)
        action.update_header(self.header)
        self.assertEqual(self.header.measurement_scheme, 'angle- and energy-dispersive')

        self.d.create_data()
        new_events = append_fields(self.d.data.events, [('lamda', np.float64), ('delta', np.float64)])
        new_events.lamda = 5.0
        new_events.delta = 0.0
        self.d.data.events = new_events
        action = CalculateQ(IncidentAngle.mu)
        action.perform_action(self.d)
        self.assertIn('qz', self.d.data.events.dtype.names)
        self.assertIn('qx', self.d.data.events.dtype.names)
        action.update_header(self.header)
        self.assertEqual(self.header.measurement_scheme, 'energy-dispersive')

        self.d.create_data()
        new_events = append_fields(self.d.data.events, [('lamda', np.float64), ('delta', np.float64)])
        new_events.lamda = 5.0
        new_events.delta = 0.0
        self.d.data.events = new_events
        action = CalculateQ(IncidentAngle.nu)
        action.perform_action(self.d)
        self.assertIn('qz', self.d.data.events.dtype.names)
        self.assertNotIn('qx', self.d.data.events.dtype.names)
        action.update_header(self.header)
        self.assertEqual(self.header.measurement_scheme, 'energy-dispersive')

    def test_filter_qz_range(self):
        action = FilterQzRange((0.1, 0.2))
        with self.assertRaises(ValueError):
            action.perform_action(self.d)

        self.d.data.events = self.d.data.events[:5].copy()
        new_events = append_fields(self.d.data.events, [('qz', np.float64)])
        new_events.qz = np.array([0.05, 0.1, 0.15, 0.2, 0.25])
        self.d.data.events = new_events
        action.perform_action(self.d)
        np.testing.assert_array_equal(
            self.d.data.events.mask,
            np.array([1, 0, 0, 0, 1], dtype=np.int32) * EVENT_BITMASKS['qRange']
        )

    def test_filter_by_log(self):
        action = FilterByLog("test_log==0") | ApplyMask()
        class LogWarnError(Exception):
            ...
        def warn_raise(*args, **kwargs):
            raise LogWarnError()
        _orig_warn = logging.warning
        try:
            logging.warning = warn_raise
            with self.assertRaises(LogWarnError):
                action.perform_action(self.d)
        finally:
            logging.warning = _orig_warn

        self._extract_walltime()

        test_log = np.recarray(shape=(2,), dtype=np.dtype([('value', np.int32),
                                                           ('time', np.int64)]))
        test_log.time = [-5, self.d.data.pulses.time[100]+123]
        test_log.value = [0, 1]
        self.d.data.device_logs['test_log'] = test_log
        action.perform_action(self.d)
        self.assertEqual(self.d.data.pulses.shape[0], 101)

    def test_filter_by_log_switchpulse(self):
        action = FilterByLog("!test_log==0") | ApplyMask()
        self._extract_walltime()

        test_log = np.recarray(shape=(2,), dtype=np.dtype([('value', np.int32),
                                                           ('time', np.int64)]))
        test_log.time = [-5, self.d.data.pulses.time[100]+123]
        test_log.value = [0, 1]
        self.d.data.device_logs['test_log'] = test_log
        self.d.data.device_logs['check_log'] = test_log.copy()
        action.perform_action(self.d)
        self.assertEqual(self.d.data.pulses.shape[0], 100)
        np.testing.assert_array_equal(
                self.d.data.device_logs['test_log'],
                self.d.data.device_logs['check_log'],
                )


"""
Specify the data type and protocol used for event datasets.
"""
from typing import Dict, List, Optional, Protocol, Tuple
from dataclasses import dataclass, field
from .header import Header
from abc import ABC, abstractmethod
from hashlib import sha256
import  numpy as np
import logging

@dataclass
class AmorGeometry:
    mu:float
    nu:float
    kap:float
    kad:float
    div:float

    chopperSeparation: float
    detectorDistance: float
    chopperDetectorDistance: float

@dataclass
class AmorTiming:
    ch1TriggerPhase: float
    ch2TriggerPhase: float
    chopperSpeed: float
    chopperPhase: float
    tau: float

# Structured datatypes used for event streams
EVENT_TYPE = np.dtype([('tof', np.float64), ('pixelID', np.uint32), ('mask', np.int32)])
PACKET_TYPE = np.dtype([('start_index', np.uint32), ('time', np.int64)])
PULSE_TYPE = np.dtype([('time', np.int64), ('monitor', np.float32)])
PC_TYPE = np.dtype([('current', np.float32), ('time', np.int64)])
LOG_TYPE = np.dtype([('value', np.float32), ('time', np.int64)])

# define the bitmask for individual event filters
EVENT_BITMASKS = {
    'MonitorThreshold': 1,
    'StrangeTimes': 2,
    'yRange': 4,
    'LamdaRange': 8,
    'qRange': 16,
    }

def append_fields(input: np.recarray, new_fields: List[Tuple[str, np.dtype]]):
    # add one ore more fields to a recarray, numpy functions seems to fail
    flds = [(name, dtypei[0]) for name, dtypei in input.dtype.fields.items()]
    flds += new_fields
    output = np.recarray(len(input), dtype=flds)

    for field in input.dtype.fields.keys():
        output[field] = input[field]
    return output

@dataclass
class AmorEventStream:
    events: np.recarray # EVENT_TYPE
    packets: np.recarray # PACKET_TYPE
    pulses: np.recarray  # PULSE_TYPE
    proton_current: np.recarray # PC_TYPE
    device_logs: Dict[str, np.recarray] = field(default_factory=dict) # LOG_TYPE

class EventDatasetProtocol(Protocol):
    """
    Minimal attributes a dataset needs to provide to work with EventDataAction
    """
    geometry: AmorGeometry
    timing: AmorTiming
    data: AmorEventStream

    def append(self, other):
        # Should define a way to add events from other to own
        ...

    def update_header(self, header:Header):
        # update a header with the information read from file
        ...

class EventDataAction(ABC):
    """
    Abstract base class used for actions applied to an EventDatasetProtocol based objects.
    Each action can optionally modify the header information.
    Actions can be combined using the pipe operator | (OR).
    """

    def __call__(self, dataset: EventDatasetProtocol)->None:
        logging.debug(f"        Enter action {self.__class__.__name__} on {dataset!r}")
        self.perform_action(dataset)

    @abstractmethod
    def perform_action(self, dataset: EventDatasetProtocol)->None: ...

    def update_header(self, header:Header)->None:
        if hasattr(self, 'action_name'):
            header.reduction.corrections.append(getattr(self, 'action_name'))

    def __or__(self, other:'EventDataAction')->'CombinedAction':
        return CombinedAction([self, other])

    def __repr__(self):
        output = self.__class__.__name__+'('
        for key,value in self.__dict__.items():
            output += f'{key}={value}, '
        return output.rstrip(', ')+')'

    def action_hash(self)->bytes:
        # generate a unique hash that encodes this action with its configuration parameters
        mh = sha256()
        mh.update(self.__class__.__name__.encode())
        for key,value in sorted(self.__dict__.items()):
            mh.update(repr(value).encode())
        return mh.hexdigest()

class CombinedAction(EventDataAction):
    """
    Used to perform multiple actions in one call. Stores a sequence of actions
    that are then performed individually one after the other.
    """
    def __init__(self, actions: List[EventDataAction]) -> None:
        self._actions = actions

    def perform_action(self, dataset: EventDatasetProtocol)->None:
        for action in self._actions:
            action(dataset)

    def update_header(self, header:Header)->None:
        for action in self._actions:
            action.update_header(header)

    def __or__(self, other:'EventDataAction')->'CombinedAction':
        return CombinedAction(self._actions+[other])

    def __repr__(self):
        output = repr(self._actions[0])
        for ai in self._actions[1:]:
            output += ' | '+repr(ai)
        return output

    def action_hash(self)->bytes:
        mh = sha256()
        for action in self._actions:
            mh.update(action.action_hash().encode())
        return mh.hexdigest()



"""
Events 2 histogram, quick reduction of single file to display during experiment.
Can be used as a live preview with automatic update when files are modified.
"""

import logging
import os

from time import sleep
from .kafka_events import KafkaEventData
from .header import Header
from .options import E2HConfig
from . import event_handling as eh, event_analysis as ea
from .projection import  TofZProjection,  YZProjection
from .kafka_serializer import ESSSerializer


class KafkaReduction:
    config: E2HConfig
    header: Header
    event_actions: eh.EventDataAction

    _last_mtime = 0.
    proj_yz: YZProjection
    proj_tofz = TofZProjection

    def __init__(self, config: E2HConfig):
        self.config = config

        self.header = Header()
        self.event_data = KafkaEventData()
        self.event_data.start()

        self.prepare_actions()

    def prepare_actions(self):
        """
        Does not do any actual reduction.
        """
        # Actions on datasets not used for normalization
        self.event_actions = eh.ApplyPhaseOffset(self.config.experiment.chopperPhaseOffset)
        self.event_actions |= eh.CorrectChopperPhase()
        self.event_actions |= ea.MergeFrames()
        self.event_actions |= eh.ApplyMask()

    def reduce(self):
        self.create_projections()
        self.read_data()
        self.add_data()

        self.serializer = ESSSerializer()
        self.serializer.start_command_thread()

        self.loop()

    def create_projections(self):
        self.proj_yz = YZProjection()
        self.proj_tofz = TofZProjection(self.event_data.timing.tau, foldback=True, combine=2)

    def read_data(self):
        # make sure the first events have arrived before starting analysis
        self.event_data.new_events.wait()
        self.dataset = self.event_data.get_events()
        self.event_actions(self.dataset)


    def add_data(self):
        self.monitor = self.dataset.monitor
        self.proj_yz.project(self.dataset, monitor=self.monitor)
        self.proj_tofz.project(self.dataset, monitor=self.monitor)

    def loop(self):
        self.wait_for = self.serializer.new_count_started
        while True:
            try:
                self.update()
                self.wait_for.wait(1.0)
            except KeyboardInterrupt:
                self.event_data.stop_event.set()
                self.event_data.join()
                self.serializer.end_command_thread()
                return

    def update(self):
        if self.serializer.new_count_started.is_set():
            logging.warning('Start new count, clearing event data')
            self.wait_for = self.serializer.count_stopped
            self.event_data.restart()
            self.serializer.new_count_started.clear()
            self.create_projections()
            return
        elif self.serializer.count_stopped.is_set() and not self.event_data.stop_counting.is_set():
            return self.finish_count()
        try:
            update_data = self.event_data.get_events()
        except EOFError:
            return
        logging.info("    updating with new data")

        self.event_actions(update_data)
        self.dataset=update_data
        self.monitor = self.dataset.monitor
        self.proj_yz.project(update_data, self.monitor)
        self.proj_tofz.project(update_data, self.monitor)

        self.serializer.send(self.proj_yz)
        self.serializer.send(self.proj_tofz)

    def finish_count(self):
        logging.debug("    stop event set, hold event collection and send final results")
        self.wait_for = self.serializer.new_count_started
        self.event_data.stop_counting.set()

        try:
            update_data = self.event_data.get_events()
        except EOFError:
            pass
        else:
            self.event_actions(update_data)
            self.dataset = update_data
            self.monitor = self.dataset.monitor
            self.proj_yz.project(update_data, self.monitor)
            self.proj_tofz.project(update_data, self.monitor)

        logging.warning(f'  stop counting, total events {int(self.proj_tofz.data.cts.sum())}')

        self.serializer.send(self.proj_yz, final=True)
        self.serializer.send(self.proj_tofz, final=True)



"""
Allows to send eos projections to Kafka using ESS histogram serialization.

For histogram_h01 the message is build using:

hist = {
    "source": "some_source",
    "timestamp": 123456,
    "current_shape": [2, 5],
    "dim_metadata": [
        {
            "length": 2,
            "unit": "a",
            "label": "x",
            "bin_boundaries": np.array([10, 11, 12]),
        },
        {
            "length": 5,
            "unit": "b",
            "label": "y",
            "bin_boundaries": np.array([0, 1, 2, 3, 4, 5]),
        },
    ],
    "last_metadata_timestamp": 123456,
    "data": np.array([[1, 2, 3, 4, 5], [6, 7, 8, 9, 10]]),
    "errors": np.array([[5, 4, 3, 2, 1], [10, 9, 8, 7, 6]]),
    "info": "info_string",
}
"""
import logging
from typing import List, Tuple, Union
from threading import Thread, Event

import numpy as np
import json
from time import time
from dataclasses import dataclass, asdict
from streaming_data_types import histogram_hs01
from confluent_kafka import Producer, Consumer

from uuid import uuid4

from .projection import TofZProjection, YZProjection

KAFKA_BROKER = 'linkafka01.psi.ch:9092'
KAFKA_TOPICS = {
    'histogram': 'amor_histograms',
    'response': 'amor_histResponse',
    'command':  'amor_histCommands'
    }

def ktime():
    return int(time()*1_000)

@dataclass
class DimMetadata:
    length: int
    unit: str
    label: str
    bin_boundaries: np.ndarray

@dataclass
class HistogramMessage:
    source: str
    timestamp: int
    current_shape: Tuple[int, int]
    dim_metadata: Tuple[DimMetadata, DimMetadata]
    last_metadata_timestamp: int
    data: np.ndarray
    errors: np.ndarray
    info: str

    def serialize(self):
        return histogram_hs01.serialise_hs01(asdict(self))

@dataclass
class CommandMessage:
    msg_id: str

    cmd=None

    @classmethod
    def get_message(cls, data):
        """
        Uses the sub-class cmd attribute to select which message to retugn
        """
        msg = dict([(ci.cmd, ci) for ci in cls.__subclasses__()])
        return msg[data['cmd']](**data)


@dataclass
class Stop(CommandMessage):
    hist_id: str
    id: str
    cmd:str = 'stop'

@dataclass
class HistogramConfig:
    id: str
    type: str
    data_brokers: List[str]
    topic: str
    data_topics: List[str]
    tof_range: Tuple[float, float]
    det_range: Tuple[int, int]
    num_bins: int
    width: int
    height: int
    left_edges: list
    source: str

@dataclass
class ConfigureHistogram(CommandMessage):
    histograms: List[HistogramConfig]
    start: int
    cmd:str = 'config'

    def __post_init__(self):
        self.histograms = [HistogramConfig(**cfg) for cfg in self.histograms]


class ESSSerializer:

    def __init__(self):
        self.producer = Producer({
            'bootstrap.servers': KAFKA_BROKER,
            'message.max.bytes': 4_000_000,
            })
        self.consumer = Consumer({
            'bootstrap.servers': KAFKA_BROKER,
            "group.id": uuid4(),
            "default.topic.config": {"auto.offset.reset": "latest"},
            })
        self._active_histogram_yz = None
        self._active_histogram_tofz = None
        self.new_count_started = Event()
        self.count_stopped = Event()

        self.consumer.subscribe([KAFKA_TOPICS['command']])

    def process_message(self, message):
        if message.error():
            logging.error("Command Consumer Error: %s", message.error())
        else:
            command = json.loads(message.value().decode())
            try:
                command = CommandMessage.get_message(command)
            except Exception:
                logging.error(f'Could not interpret message: \n{command}', exc_info=True)
                return
            logging.info(command)
            resp = json.dumps({
                "msg_id":   getattr(command, "id", None) or command.msg_id,
                "response": "ACK",
                "message":  ""
                })
            self.producer.produce(
                    topic=KAFKA_TOPICS['response'],
                    value=resp
                    )
            self.producer.flush()
            if isinstance(command, Stop):
                if command.hist_id in [self._active_histogram_yz, self._active_histogram_tofz]:
                    self.count_stopped.set()
                else:
                    return
            elif isinstance(command, ConfigureHistogram):
                for hist in command.histograms:
                    if hist.topic == KAFKA_TOPICS['histogram']+'_YZ':
                        self._active_histogram_yz = hist.id
                        logging.debug(f"   histogram data_topic: {hist.data_topics}")
                        self._start = command.start
                        self.count_stopped.clear()
                        self.new_count_started.set()
                    if hist.topic == KAFKA_TOPICS['histogram']+'_TofZ':
                        self._active_histogram_tofz = hist.id

    def receive(self, timeout=5):
        rec = self.consumer.poll(timeout)
        if rec is not None:
            self.process_message(rec)
            return True
        else:
            return False

    def receive_loop(self):
        while not self._stop_receiving.is_set():
            try:
                self.receive()
            except Exception:
                logging.error("Exception while receiving", exc_info=True)

    def start_command_thread(self):
        self._stop_receiving = Event()
        self._command_thread = Thread(target=self.receive_loop)
        self._command_thread.start()

    def end_command_thread(self, event=None):
        self._stop_receiving.set()
        self._command_thread.join()

    def acked(self, err, msg):
        # We need to have callback to produce-method to catch server errors
        if err is not None:
            logging.warning("Failed to deliver message: %s: %s" % (str(msg), str(err)))
        else:
            logging.debug("Message produced: %s" % (str(msg)))

    def send(self, proj: Union[YZProjection, TofZProjection], final=False):
        if final:
            state = 'FINISHED'
        else:
            state = 'COUNTING'
        if isinstance(proj, YZProjection):
            if self._active_histogram_yz is None:
                return
            suffix = 'YZ'
            message = HistogramMessage(
                source='amor-eos',
                timestamp=ktime(),
                current_shape=(proj.y.shape[0]-1, proj.z.shape[0]-1),
                dim_metadata=(
                    DimMetadata(
                            length=proj.y.shape[0]-1,
                            unit="pixel",
                            label="Y",
                            bin_boundaries=proj.y,
                            ),
                    DimMetadata(
                            length=proj.z.shape[0]-1,
                            unit="pixel",
                            label="Z",
                            bin_boundaries=proj.z,
                            )
                ),
                last_metadata_timestamp=0,
                data=proj.data.cts,
                errors=np.sqrt(proj.data.cts),
                info=json.dumps({
                    "start": self._start,
                    "state": state,
                    "num events": proj.data.cts.sum()
                })
                )
            logging.info(f"   {state}: Sending {proj.data.cts.sum()} events to Nicos")
        elif isinstance(proj, TofZProjection):
            if self._active_histogram_tofz is None:
                return
            suffix = 'TofZ'
            message = HistogramMessage(
                source='amor-eos',
                timestamp=ktime(),
                current_shape=(proj.tof.shape[0]-1, proj.z.shape[0]-1),
                dim_metadata=(
                    DimMetadata(
                            length=proj.tof.shape[0]-1,
                            unit="ms",
                            label="ToF",
                            bin_boundaries=proj.tof,
                            ),
                    DimMetadata(
                            length=proj.z.shape[0]-1,
                            unit="pixel",
                            label="Z",
                            bin_boundaries=proj.z,
                            ),
                ),
                last_metadata_timestamp=0,
                data=proj.data.cts,
                errors=np.sqrt(proj.data.cts),
                info=json.dumps({
                    "start": self._start,
                    "state": state,
                    "num events": proj.data.I.sum()
                })
                )
        else:
            raise NotImplementedError(f"Histogram for {proj.__class__.__name__} not implemented")

        self.producer.produce(value=message.serialize(),
                              topic=KAFKA_TOPICS['histogram']+'_'+suffix,
                              callback=self.acked)
        self.producer.flush()



"""
Collect AMOR detector events send via Kafka.
"""

import logging
import numpy as np
from threading import Thread, Event
from time import time

from .event_data_types import AmorGeometry, AmorTiming, AmorEventStream, PACKET_TYPE, EVENT_TYPE, PULSE_TYPE, PC_TYPE

from uuid import uuid4
from streaming_data_types.eventdata_ev44 import EventData
from streaming_data_types.logdata_f144 import ExtractedLogData
from streaming_data_types import deserialise_f144, deserialise_ev44
from confluent_kafka import Consumer

from .header import Header


try:
    from streaming_data_types.utils import get_schema
except ImportError:
    from streaming_data_types.utils import _get_schema as get_schema


KAFKA_BROKER = 'linkafka01.psi.ch:9092'
AMOR_EVENTS = 'amor_detector'
AMOR_NICOS = 'amor_nicosForwarder'

class KafkaFrozenData:
    """
    Represents event stream data from Kafka at a given time.
    Will be returned by KafkaEventData to be use in conjunction
    with data processing and projections.

    Implements EventDatasetProtocol
    """
    geometry: AmorGeometry
    timing: AmorTiming
    data: AmorEventStream

    def __init__(self, geometry, timing, data, monitor=1.):
        self.geometry = geometry
        self.timing = timing
        self.data = data
        self.monitor = monitor

    def append(self, other):
        raise NotImplementedError("can't append live datastream to other event data")

    def update_header(self, header:Header):
        # maybe makes sense later, but for now just used for live vizualization
        ...

class KafkaEventData(Thread):
    """
    Read Nicos information and events from Kafka. Creates a background
    thread that listens to Kafka events and converts them to eos compatible information.
    """
    geometry: AmorGeometry
    timing: AmorTiming
    events: np.recarray

    def __init__(self):
        self.stop_event = Event()
        self.stop_counting = Event()
        self.new_events = Event()
        self.last_read = 0
        self.last_read_time = 0.
        self.start_time = time()
        self.consumer = Consumer(
                {'bootstrap.servers': 'linkafka01.psi.ch:9092',
                 'group.id': uuid4()})
        self.consumer.subscribe([AMOR_EVENTS, AMOR_NICOS])
        self.geometry = AmorGeometry(1.0, 2.0, 0., 0., 1.5, 10.0, 4.0, 10.0)
        self.timing = AmorTiming(0., 0., 500., 0., 30./500.)
        # create empty dataset
        self.events = np.recarray(0, dtype=EVENT_TYPE)
        super().__init__()

    def run(self):
        while not self.stop_event.is_set():
            messages = self.consumer.consume(10, timeout=1)
            for message in messages:
                self.process_message(message)

    def process_message(self, message):
        if message.error():
            logging.info(f"  received Kafka message with error: {message.error()}")
            return
        schema = get_schema(message.value())
        if message.topic()==AMOR_EVENTS and schema=='ev44':
            events:EventData = deserialise_ev44(message.value())
            self.add_events(events)
            self.new_events.set()
            logging.debug(f'  new events {events}')
        elif message.topic()==AMOR_NICOS and schema=='f144':
            nicos_data:ExtractedLogData = deserialise_f144(message.value())
            if nicos_data.source_name in self.nicos_mapping.keys():
                logging.debug(f'  {nicos_data.source_name} = {nicos_data.value}')
                self.update_instrument(nicos_data)

    def add_events(self, events:EventData):
        """
        Add new events to the Dataset. The object keeps raw events
        and only copies the latest set to the self.data object,
        this allows to run the event processing to be performed on a "clean"
        evnet stream each time.
        """
        if self.stop_counting.is_set():
            return
        prev_size = self.events.shape[0]
        new_events = events.pixel_id.shape[0]
        self.events.resize(prev_size+new_events, refcheck=False)
        self.events.pixelID[prev_size:] = events.pixel_id
        self.events.mask[prev_size:] = 0
        self.events.tof[prev_size:] = events.time_of_flight/1.e9

    nicos_mapping = {
        'mu': ('geometry', 'mu'),
        'nu':  ('geometry', 'nu'),
        'kappa':  ('geometry', 'kap'),
        'kappa_offset':  ('geometry', 'kad'),
        'ch1_trigger_phase':  ('timing', 'ch1TriggerPhase'),
        'ch2_trigger_phase':  ('timing', 'ch2TriggerPhase'),
        'ch2_speed':  ('timing', 'chopperSpeed'),
        'chopper_phase':  ('timing', 'chopperPhase'),
        }

    def update_instrument(self, nicos_data:ExtractedLogData):
        if nicos_data.source_name in self.nicos_mapping:
            attr, subattr = self.nicos_mapping[nicos_data.source_name]
            setattr(getattr(self, attr), subattr, nicos_data.value)
            if nicos_data.source_name=='ch2_speed':
                self.timing.tau = 30./self.timing.chopperSpeed

    def monitor(self):
        return time()-self.start_time

    def restart(self):
        # empty event buffer
        self.events = np.recarray(0, dtype=EVENT_TYPE)
        self.stop_counting.clear()
        self.last_read = 0
        self.start_time = time()
        self.new_events.clear()

    def get_events(self, total_counts=False):
        packets = np.recarray(0, dtype=PACKET_TYPE)
        pulses = np.recarray(0, dtype=PULSE_TYPE)
        pc = np.recarray(0, dtype=PC_TYPE)
        if total_counts:
            last_read = 0
        else:
            last_read = self.last_read
        if last_read>=self.events.shape[0]:
            raise EOFError("No new events arrived")
        data = AmorEventStream(self.events[last_read:].copy(), packets, pulses, pc)
        self.last_read = self.events.shape[0]
        self.new_events.clear()
        t_now = time()
        monitor = t_now-self.last_read_time
        self.last_read_time = t_now
        return KafkaFrozenData(self.geometry, self.timing, data, monitor=monitor)



import argparse

from typing import List, Type
from .options import ArgParsable


def commandLineArgs(config_items: List[Type[ArgParsable]], program_name=None, extra_args=[]):
    """
    Process command line argument.
    The type of the default values is used for conversion and validation.
    """
    msg = "eos reads data from (one or several) raw file(s) of the .hdf format, \
           performs various corrections, conversations and projections and exports\
           the resulting reflectivity in an orso-compatible format."
    clas = argparse.ArgumentParser(prog=program_name,
                                   description = msg, formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    clas.add_argument('-v', '--verbose', action='count', default=0)

    clas_groups = {}

    all_arguments = []
    for cls in config_items:
        all_arguments += cls.get_commandline_parameters()

    all_arguments.sort() # parameters are sorted alphabetically, unless they have higher priority
    for cpc in all_arguments:
        if not cpc.group in clas_groups:
            clas_groups[cpc.group] = clas.add_argument_group(cpc.group)
        if cpc.short_form:
            clas_groups[cpc.group].add_argument(
                    f'-{cpc.short_form}', f'--{cpc.argument}', **cpc.add_argument_args
                    )
        else:
            clas_groups[cpc.group].add_argument(
                    f'--{cpc.argument}', **cpc.add_argument_args
                    )

    for ma in extra_args:
        clas.add_argument(**ma)

    return clas.parse_args()
